/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <gio/gio.h>
#include <sqlite3.h>
#include <sys/time.h>
#include <sys/types.h>

#include <app_manager.h>
#include <runtime_info.h>

#include "bm_plugin_interface.h"
#include "bm_cpu_util.h"
#include "bm_log.h"

enum heart_cpu_status {
	SERVICE = 0,
	FOREG = 1,
	BACKG = 2,
};

struct app_status {
	pid_t pid;
	time_t last_checkpoint;	// Need to get usage from this time
	time_t used_time;	// utime + stime
	enum heart_cpu_status last_status;
};

static struct timeval last_requested_time;

static GDBusConnection *dbus_connection;
static GHashTable *running_app_list;	// Key(appid), Val(app_status)

gint find_app_time(gconstpointer a, gconstpointer b)
{
	const char *appid = (const char *)b;
	const app_time_map_st1 *app_time = (const app_time_map_st1 *)a;

	return g_strcmp0(appid, app_time->app_id);
}

void free_atm_st1(gpointer data)
{
	app_time_map_st1 *atm = (app_time_map_st1 *)data;

	if (!atm)
		return;

	if (atm->app_id)
		free(atm->app_id);
	free(atm);
}

bool is_running_app(const char *appid)
{
	bool is_running = false;
	int ret = app_manager_is_running(appid, &is_running);
	if (ret != APP_MANAGER_ERROR_NONE) {
		_E("app_manager_is_running failed (%d)", ret);
		return false;
	}

	return is_running;
}

int deinit()
{
	ENTER;

	if (dbus_connection) {
		g_object_unref(dbus_connection);
		dbus_connection = NULL;
	}

	if (running_app_list) {
		g_hash_table_destroy(running_app_list);
		running_app_list = NULL;
	}

	EXIT;
	return BM_PLUGIN_ERROR_NONE;
}

int init()
{
	ENTER;

	GError *g_err = NULL;

	if (gettimeofday(&last_requested_time, NULL) != 0) {
		_E("gettimeofday failed : %m");
		goto failed;
	}

	dbus_connection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &g_err);
	if (!dbus_connection) {
		_E("g_bus_get_sync failed : %s", g_err->message ? g_err->message : NULL);
		g_error_free(g_err);
		goto failed;
	}

	running_app_list = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	if (!running_app_list) {
		_E("g_hash_table_new_full failed");
		goto failed;
	}

	EXIT;
	return BM_PLUGIN_ERROR_NONE;

failed:
	deinit();

	EXIT;
	return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
}

int get_feature_data(bm_data_h *handle, bm_plugin_data_type_e type)
{
	ENTER;

	int ret;

	bm_cpu_st *usage = NULL;

	struct timeval tv;

	GError *g_err = NULL;
	GDBusMessage *msg_req, *msg_reply;

	/* Sqlite */
	sqlite3 *db;
	sqlite3_stmt *stmt;
	char query[128];

	/* DB row */
	unsigned int time;
	const char *appid;
	const char *data;
	unsigned int utime, stime, pid, status;

	/* App status */
	struct app_status *app_status, *new_app_status;
	int elapsed;
	process_cpu_usage_s *cpu_usage;
	GHashTableIter iter;
	gpointer g_key, g_val;

	/* App time map */
	app_time_map_st1 *app_time_map;


	/* Check argument */
	if (!handle || type != BM_DATA_TYPE_CPU) {
		_E("Invalid argument");
		return BM_PLUGIN_ERROR_INVALID_PARAMETER;
	}

	/* Get current time */
	ret = gettimeofday(&tv, NULL);
	if (ret != 0) {
		_E("gettimeofday failed : %m");

		EXIT;
		return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
	}

	/* Request for resourced to update HEART-CPU DB */
	msg_req = g_dbus_message_new_method_call(
			"org.tizen.resourced",
			"/Org/Tizen/ResourceD/Logging",
			"org.tizen.resourced.logging",
			"SyncCpuData"
			);
	if (!msg_req) {
		if (g_err) {
			_E("g_dbus_message_new_method_call failed : %s", g_err->message ? g_err->message : NULL);
			g_error_free(g_err);
		} else
			_E("g_dbus_message_new_method_call failed");

		EXIT;
		return BM_PLUGIN_ERROR_NOT_SUPPORTED;
	}

	msg_reply = g_dbus_connection_send_message_with_reply_sync(dbus_connection, msg_req, G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &g_err);
	if (msg_reply)
		g_object_unref(msg_reply);
	else {
		if (g_err) {
			_W("g_dbus_connection_send_message_with_reply_sync failed : %s", g_err->message ? g_err->message : NULL);
			g_error_free(g_err);
		} else
			_W("g_dbus_connection_send_message_with_reply_sync failed");
	}

	/* Open HEART-CPU DB */
	ret = sqlite3_open("/opt/usr/home/owner/.applications/dbspace/.resourced-heart-default.db", &db);
	if (ret != SQLITE_OK) {
		_E("sqlite3_open failed (%d)", ret);

		EXIT;
		return BM_PLUGIN_ERROR_NO_DATA;
	}

	/* Make usage data structure */
	usage = (bm_cpu_st *)calloc(1, sizeof(bm_cpu_st));
	if (!usage) {
		_E("calloc failed");

		EXIT;
		return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
	}

	/* Read data from HEART-CPU DB */
	_I("Gather CPU usage : %ld ~ %ld", last_requested_time.tv_sec, tv.tv_sec);
	snprintf(query, 128, "SELECT time,appid,data FROM cpu WHERE time >= %ld", last_requested_time.tv_sec);
	ret = sqlite3_prepare_v2(db, query, -1, &stmt, 0);
	while (sqlite3_step(stmt) == SQLITE_ROW) {
		time = sqlite3_column_int(stmt, 0);
		appid = (const char *)sqlite3_column_text(stmt, 1);
		data = (const char *)sqlite3_column_text(stmt, 2);
		sscanf(data, "%u %u %u %u", &utime, &stime, &pid, &status);

		//_D("time(%u), appid(%s), utime(%u), stime(%u), pid(%u), status(%u)", time, appid, utime, stime, pid, status);

		/* Ignore terminated app */
		if (!is_running_app(appid)) {
			_D("%s is already terminated", appid);
			g_hash_table_remove(running_app_list, appid);
			continue;
		}

		app_status = g_hash_table_lookup(running_app_list, appid);

		/* Insert newly launched app in the running app list */
		if (!app_status) {
			_D("%s is newly launched", appid);
			new_app_status = malloc(sizeof(struct app_status));
			if (!new_app_status) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			new_app_status->pid = pid;
			new_app_status->last_checkpoint = time;
			new_app_status->used_time = utime + stime;
			new_app_status->last_status = status;
			g_hash_table_insert(running_app_list, g_strdup(appid), new_app_status);
			continue;
		}

		/* Initialize restarted app */
		if (app_status->pid != pid) {
			_D("%s is restarted", appid);
			app_status->pid = pid;
			app_status->last_checkpoint = time;
			app_status->used_time = utime + stime;
			app_status->last_status = status;
			continue;
		}

		/* Ignore CPU unused app */
		elapsed = utime + stime - app_status->used_time;
		if (elapsed <= 0) {
			_D("%s doesn't use CPU. Ignore it", appid);
			continue;
		}

		/* Accumulate used time */
		_D("%s uses CPU (+%d)", appid, elapsed);
		usage->cpuTicks += elapsed;
		switch (app_status->last_status) {
		case FOREG:
			usage->cpuTimeForeground += elapsed;
			break;
		case BACKG:
		case SERVICE:
			usage->cpuTimeBackground += elapsed;
			break;
		}

		/* Insert the app using CPU in the ATM list */
		if (!g_slist_find_custom(usage->atm_list, appid, find_app_time)) {
			app_time_map = malloc(sizeof(app_time_map_st1));
			if (!app_time_map) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->app_id = g_strdup(appid);
			if (!app_time_map->app_id) {
				_E("g_strdup failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->time = tv.tv_sec - app_status->last_checkpoint;
			if (app_time_map->time <= 0)
				app_time_map->time = 1;
			usage->atm_list = g_slist_append(usage->atm_list, app_time_map);
		}

		/* Update app status */
		app_status->used_time = utime + stime;
		app_status->last_status = status;
	}

	/* Update all running apps with current value */
	g_hash_table_iter_init(&iter, running_app_list);
	while (g_hash_table_iter_next(&iter, &g_key, &g_val)) {
		appid = (const char *)g_key;
		app_status = (struct app_status *)g_val;

		/* Ignore terminated app */
		if (!is_running_app(appid)) {
			_D("%s is already terminated", appid);
			g_hash_table_iter_remove(&iter);
			continue;
		}

		/* Get latest CPU time */
		ret = runtime_info_get_process_cpu_usage(&app_status->pid, 1, &cpu_usage);
		if (ret != RUNTIME_INFO_ERROR_NONE) {
			_E("runtime_info_get_process_cpu_usage for %u failed(%d)", app_status->pid, ret);
			continue;
		}

		app_status->last_checkpoint = tv.tv_sec;

		/* Ignore CPU unused app */
		elapsed = cpu_usage->utime + cpu_usage->stime - app_status->used_time;
		if (elapsed <= 0) {
			_D("%s doesn't use CPU. Ignore it", appid);
			continue;
		}

		/* Update CPU time */
		_D("%s uses CPU (+%d)", appid, elapsed);
		usage->cpuTicks += elapsed;
		switch (app_status->last_status) {
		case FOREG:
			usage->cpuTimeForeground += elapsed;
			break;
		case BACKG:
		case SERVICE:
			usage->cpuTimeBackground += elapsed;
			break;
		}

		/* Insert the app using CPU in the ATM list */
		if (!g_slist_find_custom(usage->atm_list, appid, find_app_time)) {
			app_time_map = malloc(sizeof(app_time_map_st1));
			if (!app_time_map) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->app_id = g_strdup(appid);
			if (!app_time_map->app_id) {
				_E("g_strdup failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->time = tv.tv_sec - last_requested_time.tv_sec;
			if (app_time_map->time <= 0)
				app_time_map->time = 1;
			usage->atm_list = g_slist_append(usage->atm_list, app_time_map);
		}

		app_status->used_time = cpu_usage->utime + cpu_usage->stime;

		free(cpu_usage);
	}

	/* Update last requested time */
	last_requested_time.tv_sec = tv.tv_sec;
	last_requested_time.tv_usec = tv.tv_usec;

	/* Return result */
	*handle = usage;

	_I("Succeed to get CPU usage");
	ret = BM_PLUGIN_ERROR_NONE;
	goto clean_db;

clean_atm:
	g_slist_free_full(usage->atm_list, free_atm_st1);

clean_db:
	sqlite3_finalize(stmt);
	sqlite3_close(db);

	EXIT;
	return ret;
}

static bm_api_st cpu_api = {
	.init = init,
	.deinit = deinit,
	.get_feature_data = get_feature_data
};

int plugin_init(bm_api_st **api)
{
	ENTER;

	*api = &cpu_api;

	EXIT;
	return 0;
}

int plugin_deinit(void)
{
	ENTER;

	EXIT;
	return 0;
}

BM_PLUGIN_ADD(BM_PLUGIN_ID_CPU, NAME, AUTHOR, VERSION, plugin_init, plugin_deinit);
